# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    Client for litecoind deamon
#    © 2018-2022 Oct - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import configparser

from fluxwallet.main import *
from fluxwallet.networks import Network
from fluxwallet.services.authproxy import AuthServiceProxy
from fluxwallet.services.baseclient import BaseClient, ClientError
from fluxwallet.transactions import BitcoinTransaction

PROVIDERNAME = "litecoind"

_logger = logging.getLogger(__name__)


class ConfigError(Exception):
    def __init__(self, msg=""):
        self.msg = msg
        _logger.info(msg)

    def __str__(self):
        return self.msg


def _read_from_config(configparser, section, value, fallback=None):
    try:
        return configparser.get(section, value)
    except Exception:
        return fallback


class LitecoindClient(BaseClient):
    """
    Class to interact with litecoind, the Litecoin deamon
    """

    @staticmethod
    def from_config(configfile=None, network="litecoin"):
        """
        Read settings from litecoind config file

        :param configfile: Path to config file. Leave empty to look in default places
        :type: str
        :param network: Litecoin mainnet or testnet. Default is litecoin mainnet
        :type: str

        :return LitecoindClient:
        """
        try:
            config = configparser.ConfigParser(strict=False)
        except TypeError:
            config = configparser.ConfigParser()
        config_fn = "litecoin.conf"
        if isinstance(network, Network):
            network = network.name
        if network == "testnet":
            config_fn = "litecoin-testnet.conf"

        cfn = None
        if not configfile:
            config_locations = [
                "~/.fluxwallet",
                "~/.litecoin",
                "~/Application Data/Litecoin",
                "~/Library/Application Support/Litecoin",
            ]
            for location in config_locations:
                cfn = Path(location, config_fn).expanduser()
                if cfn.exists():
                    break
        else:
            cfn = Path(FW_DATA_DIR, "config", configfile)

        if not cfn or not cfn.is_file():
            raise ConfigError(
                "Config file %s not found. Please install Litecoin client and specify a path to config "
                "file if path is not default. Or place a config file in .fluxwallet/litecoin.conf to "
                "reference to an external server." % cfn
            )

        try:
            config.read(cfn)
        except Exception:
            with cfn.open() as f:
                config_string = "[rpc]\n" + f.read()
            config.read_string(config_string)

        testnet = _read_from_config(config, "rpc", "testnet")
        if testnet:
            network = "testnet"
        if _read_from_config(config, "rpc", "rpcpassword") == "specify_rpc_password":
            raise ConfigError("Please update config settings in %s" % cfn)
        if network == "testnet":
            port = 19332
        else:
            port = 9332
        port = _read_from_config(config, "rpc", "rpcport", port)
        server = "127.0.0.1"
        server = _read_from_config(config, "rpc", "rpcconnect", server)
        server = _read_from_config(config, "rpc", "bind", server)
        server = _read_from_config(config, "rpc", "externalip", server)
        url = "http://%s:%s@%s:%s" % (
            config.get("rpc", "rpcuser"),
            config.get("rpc", "rpcpassword"),
            server,
            port,
        )
        return LitecoindClient(network, url)

    def __init__(self, network="litecoin", base_url="", denominator=100000000, *args):
        """
        Open connection to litecoin node

        :param network: Litecoin mainnet or testnet. Default is litecoin mainnet
        :type: str
        :param base_url: Connection URL in format http(s)://user:password@host:port.
        :type: str
        :param denominator: Denominator for this currency. Should be always 100000000 (satoshis) for litecoin
        :type: str
        """
        if isinstance(network, Network):
            network = network.name
        if not base_url:
            bdc = self.from_config("", network)
            base_url = bdc.base_url
            network = bdc.network
        if len(base_url.split(":")) != 4:
            raise ConfigError(
                "Litecoind connection URL must be of format 'http(s)://user:password@host:port,"
                "current format is %s. Please set url in providers.json file or check litecoin config "
                "file" % base_url
            )
        if "password" in base_url:
            raise ConfigError(
                "Invalid password in litecoind provider settings. "
                "Please replace default password and set url in providers.json or litecoin.conf file"
            )
        _logger.info("Connect to litecoind")
        self.proxy = AuthServiceProxy(base_url)
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def getbalance(self, addresslist):
        balance = 0
        for address in addresslist:
            res = self.proxy.getaddressinfo(address)
            if not (res["ismine"] or res["iswatchonly"]):
                raise ClientError(
                    "Address %s not found in litceoind wallet, use 'importpubkey' or 'importaddress' to add "
                    "address to wallet." % address
                )
            txs_list = self.proxy.listunspent(0, 99999999, [address])
            for tx in txs_list:
                balance += int(tx["amount"] * self.units)
        return balance

    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        utxos = []
        res = self.proxy.getaddressinfo(address)
        if not (res["ismine"] or res["iswatchonly"]):
            raise ClientError(
                "Address %s not found in litecoind wallet, use 'importpubkey' or 'importaddress' to add "
                "address to wallet." % address
            )

        txs_list = self.proxy.listunspent(0, 99999999, [address])
        blockcount = self.blockcount()
        for tx in sorted(txs_list, key=lambda x: x["confirmations"], reverse=True):
            utxos.append(
                {
                    "address": tx["address"],
                    "txid": tx["txid"],
                    "confirmations": tx["confirmations"],
                    "output_n": tx["vout"],
                    "input_n": -1,
                    "block_height": blockcount - tx["confirmations"] + 1,
                    "fee": None,
                    "size": 0,
                    "value": int(tx["amount"] * self.units),
                    "script": tx["scriptPubKey"],
                    "date": None,
                }
            )
            if tx["txid"] == after_txid:
                utxos = []

        return utxos

    def _parse_transaction(self, tx, block_height=None, get_input_values=True):
        t = BitcoinTransaction.parse_hex(
            tx["hex"], strict=self.strict, network=self.network
        )
        t.confirmations = tx.get("confirmations")
        t.block_hash = tx.get("blockhash")
        t.status = "unconfirmed"
        for i in t.inputs:
            if i.prev_txid == b"\x00" * 32:
                i.script_type = "coinbase"
                continue
            if get_input_values:
                txi = self.proxy.getrawtransaction(i.prev_txid.hex(), 1)
                i.value = int(
                    round(
                        float(txi["vout"][i.output_n_int]["value"])
                        / self.network.denominator
                    )
                )
        for o in t.outputs:
            o.spent = None

        if not block_height and t.block_hash:
            block_height = self.proxy.getblock(t.block_hash, 1)["height"]
        t.block_height = block_height
        if not t.confirmations and block_height is not None:
            if not self.latest_block:
                self.latest_block = self.blockcount()
            t.confirmations = (self.latest_block - block_height) + 1
        if t.confirmations or block_height:
            t.status = "confirmed"
            t.verified = True
        t.version = tx["version"].to_bytes(4, "big")
        t.version_int = tx["version"]
        t.date = None if "time" not in tx else datetime.utcfromtimestamp(tx["time"])
        t.update_totals()
        return t

    def gettransaction(self, txid):
        tx_raw = self.proxy.getrawtransaction(txid, 1)
        return self._parse_transaction(tx_raw)

    def gettransactions(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        MAX_WALLET_TRANSACTIONS = 1000
        txs = []
        res = self.proxy.getaddressinfo(address)
        if not (res["ismine"] or res["iswatchonly"]):
            raise ClientError(
                "Address %s not found in bitcoind wallet, use 'importpubkey' or 'importaddress' to add "
                "address to wallet." % address
            )
        txs_list = self.proxy.listtransactions("*", MAX_WALLET_TRANSACTIONS, 0, True)
        if len(txs_list) >= MAX_WALLET_TRANSACTIONS:
            raise ClientError(
                "Bitcoind wallet contains too many transactions %d, use other service provider for this "
                "wallet" % MAX_WALLET_TRANSACTIONS
            )
        txids = list(
            set(
                [
                    (tx["txid"], tx["blockheight"])
                    for tx in txs_list
                    if tx["address"] == address
                ]
            )
        )
        for txid, blockheight in txids:
            tx_raw = self.proxy.getrawtransaction(txid, 1)
            t = self._parse_transaction(tx_raw, blockheight)
            txs.append(t)
            if txid == after_txid:
                txs = []
        return txs

    def getrawtransaction(self, txid):
        res = self.proxy.getrawtransaction(txid)
        return res

    def sendrawtransaction(self, rawtx):
        res = self.proxy.sendrawtransaction(rawtx)
        return {"txid": res, "response_dict": res}

    def estimatefee(self, blocks):
        pres = ""
        try:
            pres = self.proxy.estimatesmartfee(blocks)
            res = pres["feerate"]
        except KeyError as e:
            _logger.info("litecoind error: %s, %s" % (e, pres))
            res = self.proxy.estimatefee(blocks)
        return int(res * self.units)

    def blockcount(self):
        return self.proxy.getblockcount()

    def mempool(self, txid=""):
        txids = self.proxy.getrawmempool()
        if not txid:
            return txids
        elif txid in txids:
            return [txid]
        return []

    def getblock(self, blockid, parse_transactions=True, page=1, limit=None):
        if isinstance(blockid, int) or len(blockid) < 10:
            blockid = self.proxy.getblockhash(int(blockid))
        if not limit:
            limit = 99999

        txs = []
        if parse_transactions:
            bd = self.proxy.getblock(blockid, 2)
            for tx in bd["tx"][(page - 1) * limit : page * limit]:
                tx["time"] = bd["time"]
                tx["blockhash"] = bd["hash"]
                txs.append(
                    self._parse_transaction(
                        tx, block_height=bd["height"], get_input_values=True
                    )
                )
        else:
            bd = self.proxy.getblock(blockid, 1)
            txs = bd["tx"]

        block = {
            "bits": int(bd["bits"], 16),
            "depth": bd["confirmations"],
            "block_hash": bd["hash"],
            "height": bd["height"],
            "merkle_root": bd["merkleroot"],
            "nonce": bd["nonce"],
            "prev_block": None
            if "previousblockhash" not in bd
            else bd["previousblockhash"],
            "time": bd["time"],
            "tx_count": bd["nTx"],
            "txs": txs,
            "version": bd["version"],
            "page": page,
            "pages": None,
            "limit": limit,
        }
        return block

    def getrawblock(self, blockid):
        if isinstance(blockid, int):
            blockid = self.proxy.getblockhash(blockid)
        return self.proxy.getblock(blockid, 0)

    def isspent(self, txid, index):
        res = self.proxy.gettxout(txid, index)
        if not res:
            return 1
        return 0

    def getinfo(self):
        info = self.proxy.getmininginfo()
        return {
            "blockcount": info["blocks"],
            "chain": info["chain"],
            "difficulty": int(info["difficulty"]),
            "hashrate": int(info["networkhashps"]),
            "mempool_size": int(info["pooledtx"]),
        }


if __name__ == "__main__":
    #
    # SOME EXAMPLES
    #

    from pprint import pprint

    # 1. Connect by specifying connection URL
    # base_url = 'http://litecoin:passwd@host:9432'
    # bdc = LitecoindClient(base_url=base_url)
    # 2. Or connect using default settings or settings from config file
    client = LitecoindClient()

    print("\n=== SERVERINFO ===")
    pprint(client.proxy.getnetworkinfo())

    print("\n=== Best Block ===")
    blockhash = client.proxy.getbestblockhash()
    bestblock = client.proxy.getblock(blockhash)
    bestblock["tx"] = "..." + str(len(bestblock["tx"])) + " transactions..."
    pprint(bestblock)

    print("\n=== Mempool ===")
    rmp = client.proxy.getrawmempool()
    pprint(rmp[:25])
    print("... truncated ...")
    print("Mempool Size %d" % len(rmp))

    print("\n=== Raw BitcoinTransaction by txid ===")
    t = client.getrawtransaction(
        "fa3906a4219078364372d0e2715f93e822edd0b47ce146c71ba7ba57179b50f6"
    )
    pprint(t)

    print("\n=== Current network fees ===")
    t = client.estimatefee(5)
    pprint(t)



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    Blocksmurfer client
#    © 2020 Januari - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime

from fluxwallet.main import MAX_TRANSACTIONS
from fluxwallet.services.baseclient import BaseClient, ClientError
from fluxwallet.transactions import BitcoinTransaction

PROVIDERNAME = "blocksmurfer"


_logger = logging.getLogger(__name__)


class BlocksmurferClient(BaseClient):
    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def compose_request(
        self,
        function,
        parameter="",
        parameter2="",
        variables=None,
        post_data="",
        method="get",
    ):
        url_path = function
        if parameter:
            url_path += "/" + str(parameter)
        if parameter2:
            url_path += "/" + str(parameter2)
        if variables is None:
            variables = {}
        if self.api_key:
            variables.update({"token": self.api_key})
        return self.request(url_path, variables, method, post_data=post_data)

    def getbalance(self, addresslist):
        balance = 0
        for address in addresslist:
            res = self.compose_request("address_balance", address)
            balance += res["balance"]
        return balance

    # TODO: fix blocksmurfer api
    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        res = self.compose_request(
            "utxos", address, variables={"after_txid": after_txid}
        )
        self.latest_block = (
            self.blockcount() if not self.latest_block else self.latest_block
        )
        utxos = []
        for u in res:
            block_height = None if not u["block_height"] else u["block_height"]
            confirmations = u["confirmations"]
            if block_height and not confirmations:
                confirmations = self.latest_block - block_height
            utxos.append(
                {
                    "address": address,
                    "txid": u["txid"],
                    "confirmations": confirmations,
                    "output_n": u["output_n"],
                    "input_n": u["input_n"],
                    "block_height": block_height,
                    "fee": u["fee"],
                    "size": u["size"],
                    "value": u["value"],
                    "script": u["script"],
                    "date": datetime.strptime(u["date"][:19], "%Y-%m-%dT%H:%M:%S"),
                }
            )
        return utxos[:limit]

    def _parse_transaction(self, tx, block_height=None):
        block_height = block_height if not tx["block_height"] else tx["block_height"]
        confirmations = tx["confirmations"]
        if block_height and not confirmations and tx["status"] == "confirmed":
            self.latest_block = (
                self.blockcount() if not self.latest_block else self.latest_block
            )
            confirmations = self.latest_block - block_height
        # FIXME: Blocksmurfer returns 'date' or 'time', should be consistent
        tx_date = (
            None
            if not tx.get("date")
            else datetime.strptime(tx["date"], "%Y-%m-%dT%H:%M:%S")
        )
        if not tx_date and "time" in tx:
            tx_date = datetime.utcfromtimestamp(tx["time"])
        t = BitcoinTransaction(
            locktime=tx["locktime"],
            version=tx["version"],
            network=self.network,
            fee=tx["fee"],
            size=tx["size"],
            txid=tx["txid"],
            date=tx_date,
            input_total=tx["input_total"],
            output_total=tx["output_total"],
            confirmations=confirmations,
            block_height=block_height,
            status=tx["status"],
            coinbase=tx["coinbase"],
            rawtx=bytes.fromhex(tx["raw_hex"]),
            witness_type=tx["witness_type"],
        )
        for ti in tx["inputs"]:
            t.add_input(
                prev_txid=ti["prev_txid"],
                output_n=ti["output_n"],
                keys=ti.get("keys", []),
                index_n=ti["index_n"],
                unlocking_script=ti["script"],
                value=ti["value"],
                public_hash=bytes.fromhex(ti["public_hash"]),
                address=ti["address"],
                witness_type=ti["witness_type"],
                locktime_cltv=ti["locktime_cltv"],
                locktime_csv=ti["locktime_csv"],
                signatures=ti["signatures"],
                compressed=ti["compressed"],
                encoding=ti["encoding"],
                unlocking_script_unsigned=ti["script_code"],
                sigs_required=ti["sigs_required"],
                sequence=ti["sequence"],
                witnesses=[bytes.fromhex(w) for w in ti["witnesses"]],
                script_type=ti["script_type"],
                strict=self.strict,
            )
        for to in tx["outputs"]:
            t.add_output(
                value=to["value"],
                address=to["address"],
                public_hash=to["public_hash"],
                lock_script=to["script"],
                spent=to["spent"],
                strict=self.strict,
            )
        t.update_totals()
        return t

    def gettransaction(self, txid, block_count=None):
        tx = self.compose_request("transaction", txid)
        return self._parse_transaction(tx, block_count)

    def gettransactions(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        prtxs = []
        while True:
            txs = self.compose_request(
                "transactions", address, variables={"after_txid": after_txid}
            )
            prtxs += txs
            if not txs or len(txs) < limit:
                break
            after_txid = txs[-1:][0]["txid"]
        txs = []
        for tx in prtxs:
            t = self._parse_transaction(tx)
            if t:
                txs.append(t)
        return txs[:limit]

    def getrawtransaction(self, txid):
        tx = self.compose_request("transaction", txid, variables={"raw": True})
        return tx["raw_hex"]

    def sendrawtransaction(self, rawtx):
        res = self.compose_request(
            "transaction_broadcast", post_data=rawtx, method="post"
        )
        return {"txid": res["txid"], "response_dict": res}

    def estimatefee(self, blocks):
        variables = {"blocks": str(blocks)}
        res = self.compose_request("fees", variables=variables)
        return res["estimated_fee_sat_kb"]

    def blockcount(self):
        return self.compose_request("blockcount")["blockcount"]

    def mempool(self, txid):
        if txid:
            t = self.gettransaction(txid)
            if t and not t.confirmations:
                return [t.txid]
        # else:
        # return self.compose_request('mempool', 'txids')
        return []

    def getblock(self, blockid, parse_transactions, page, limit):
        variables = {
            "parse_transactions": parse_transactions,
            "page": page,
            "limit": limit,
        }
        bd = self.compose_request("block", str(blockid), variables=variables)

        txs = []
        if (
            parse_transactions
            and bd["transactions"]
            and isinstance(bd["transactions"][0], dict)
        ):
            self.latest_block = (
                self.blockcount() if not self.latest_block else self.latest_block
            )
            for tx in bd["transactions"]:
                tx["confirmations"] = bd["depth"]
                tx["time"] = bd["time"]
                tx["block_height"] = bd["height"]
                tx["block_hash"] = bd["block_hash"]
                t = self._parse_transaction(tx, self.latest_block)
                if t.txid != tx["txid"]:
                    raise ClientError(
                        "Could not parse tx %s. Different txid's" % (tx["txid"])
                    )
                txs.append(t)
        else:
            txs = bd["transactions"]

        block = {
            "bits": bd["bits"],
            "depth": bd["depth"],
            "block_hash": bd["block_hash"],
            "height": bd["height"],
            "merkle_root": bd["merkle_root"],
            "nonce": bd["nonce"],
            "prev_block": bd["prev_block"],
            "time": bd["time"],
            "tx_count": bd["tx_count"],
            "txs": txs,
            "version": bd["version"],
            "page": page,
            "pages": None
            if not limit
            else int(bd["tx_count"] // limit) + (bd["tx_count"] % limit > 0),
            "limit": limit,
        }
        return block

    def getrawblock(self, blockid):
        res = self.compose_request("rawblock", blockid)
        return res

    def isspent(self, txid, output_n):
        res = self.compose_request("isspent", txid, str(output_n))
        return 1 if res["spent"] else 0

    def getinfo(self):
        res = self.compose_request("")
        info = {
            k: v
            for k, v in res.items()
            if k in ["chain", "blockcount", "hashrate", "mempool_size", "difficulty"]
        }
        return info



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    BlockCypher client
#    © 2017-2019 July - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime, timezone

from fluxwallet.main import MAX_TRANSACTIONS
from fluxwallet.services.baseclient import BaseClient, ClientError
from fluxwallet.transactions import BitcoinTransaction

PROVIDERNAME = "blockcypher"

_logger = logging.getLogger(__name__)


class BlockCypher(BaseClient):
    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def compose_request(
        self, function, data, parameter="", variables=None, method="get"
    ):
        url_path = function + "/" + data
        if parameter:
            url_path += "/" + parameter
        if variables is None:
            variables = {}
        if self.api_key:
            variables.update({"token": self.api_key})
        return self.request(url_path, variables, method)

    def getbalance(self, addresslist):
        addresslist = self._addresslist_convert(addresslist)
        addresses = ";".join([a.address for a in addresslist])
        res = self.compose_request("addrs", addresses, "balance")
        balance = 0.0
        if not isinstance(res, list):
            res = [res]
        for rec in res:
            balance += float(rec["final_balance"])
        return int(balance * self.units)

    # Disabled: Invalid results for https://api.blockcypher.com/v1/ltc/main/addrs/LVqLipGhyQ1nWtPPc8Xp3zn6JxcU1Hi8eG?unspentOnly=1&limit=2000
    # def getutxos(self, address, after_txid='', limit=MAX_TRANSACTIONS):
    #     address = self._address_convert(address)
    #     res = self.compose_request('addrs', address.address, variables={'unspentOnly': 1, 'limit': 2000})
    #     transactions = []
    #     if not isinstance(res, list):
    #         res = [res]
    #     for a in res:
    #         txrefs = a.setdefault('txrefs', []) + a.get('unconfirmed_txrefs', [])
    #         if len(txrefs) > 500:
    #             _logger.warning("BlockCypher: Large number of transactions for address %s, "
    #                             "BitcoinTransaction list may be incomplete" % address)
    #         for tx in txrefs:
    #             if tx['tx_hash'] == after_txid:
    #                 break
    #             tdate = None
    #             if 'confirmed' in tx:
    #                 try:
    #                     tdate = datetime.strptime(tx['confirmed'], "%Y-%m-%dT%H:%M:%SZ")
    #                 except ValueError:
    #                     tdate = datetime.strptime(tx['confirmed'], "%Y-%m-%dT%H:%M:%S.%fZ")
    #             transactions.append({
    #                 'address': address.address_orig,
    #                 'txid': tx['tx_hash'],
    #                 'confirmations': tx['confirmations'],
    #                 'output_n': tx['tx_output_n'],
    #                 'index': 0,
    #                 'value': int(round(tx['value'] * self.units, 0)),
    #                 'script': '',
    #                 'block_height': None,
    #                 'date': tdate
    #             })
    #     return transactions[::-1][:limit]

    def gettransaction(self, txid):
        tx = self.compose_request("txs", txid, variables={"includeHex": "true"})
        t = BitcoinTransaction.parse_hex(
            tx["hex"], strict=self.strict, network=self.network
        )
        if tx["confirmations"]:
            t.status = "confirmed"
            t.date = datetime.strptime(tx["confirmed"][:19], "%Y-%m-%dT%H:%M:%S")
        else:
            t.status = "unconfirmed"
        t.confirmations = tx["confirmations"]
        t.block_height = tx["block_height"] if tx["block_height"] > 0 else None
        t.fee = tx["fees"]
        t.rawtx = bytes.fromhex(tx["hex"])
        t.size = int(len(tx["hex"]) / 2)
        t.network = self.network
        t.input_total = 0
        if len(t.inputs) != len(tx["inputs"]):
            raise ClientError(
                "Invalid number of inputs provided. Raw tx: %d, blockcypher: %d"
                % (len(t.inputs), len(tx["inputs"]))
            )
        for n, i in enumerate(t.inputs):
            if not t.coinbase and not (
                tx["inputs"][n]["output_index"] == i.output_n_int
                and tx["inputs"][n]["prev_hash"] == i.prev_txid.hex()
            ):
                raise ClientError(
                    "BitcoinTransaction inputs do not match raw transaction"
                )
            if "output_value" in tx["inputs"][n]:
                if not t.coinbase:
                    i.value = tx["inputs"][n]["output_value"]
                t.input_total += i.value
        if len(t.outputs) != len(tx["outputs"]):
            raise ClientError(
                "Invalid number of outputs provided. Raw tx: %d, blockcypher: %d"
                % (len(t.outputs), len(tx["outputs"]))
            )
        for n, o in enumerate(t.outputs):
            if "spent_by" in tx["outputs"][n]:
                o.spent = True
                o.spending_txid = tx["outputs"][n]["spent_by"]
        return t

    def gettransactions(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        txs = []
        address = self._address_convert(address)
        res = self.compose_request(
            "addrs", address.address, variables={"unspentOnly": 0, "limit": 2000}
        )
        if not isinstance(res, list):
            res = [res]
        for a in res:
            txrefs = a.get("txrefs", []) + a.get("unconfirmed_txrefs", [])
            txids = []
            for t in txrefs[::-1]:
                if t["tx_hash"] not in txids:
                    txids.append(t["tx_hash"])
                if t["tx_hash"] == after_txid:
                    txids = []
            if len(txids) > 500:
                _logger.info(
                    "BlockCypher: Large number of transactions for address %s, "
                    "BitcoinTransaction list may be incomplete" % address.address_orig
                )
            for txid in txids[:limit]:
                t = self.gettransaction(txid)
                txs.append(t)
        return txs

    def getrawtransaction(self, txid):
        return self.compose_request("txs", txid, variables={"includeHex": "true"})[
            "hex"
        ]

    def sendrawtransaction(self, rawtx):
        # BlockCypher sometimes accepts transactions, but does not push them to the network :(
        if self.network.name in ["bitcoin", "litecoin"]:
            raise ClientError(
                "Avoid stuck transactions, skip usage of blockcypher provider"
            )
        res = self.compose_request(
            "txs", "push", variables={"tx": rawtx}, method="post"
        )
        return {"txid": res["tx"]["hash"], "response_dict": res}

    def estimatefee(self, blocks):
        res = self.compose_request("", "")
        if blocks <= 10:
            return res["medium_fee_per_kb"]
        else:
            return res["low_fee_per_kb"]

    def blockcount(self):
        return self.compose_request("", "")["height"]

    def mempool(self, txid):
        if txid:
            tx = self.compose_request("txs", txid)
            if tx["confirmations"] == 0:
                return [tx["hash"]]
            else:
                return []
        return False

    def getblock(self, blockid, parse_transactions, page, limit):
        if limit > 100:
            limit = 100
        bd = self.compose_request(
            "blocks",
            str(blockid),
            variables={"limit": limit, "txstart": ((page - 1) * limit)},
        )
        if parse_transactions:
            txs = []
            for txid in bd["txids"]:
                try:
                    txs.append(self.gettransaction(txid))
                except Exception as e:
                    _logger.error("Could not parse tx %s with error %s" % (txid, e))
        else:
            txs = bd["txids"]

        block = {
            "bits": bd["bits"],
            "depth": bd["depth"],
            "block_hash": bd["hash"],
            "height": bd["height"],
            "merkle_root": bd["mrkl_root"],
            "nonce": bd["nonce"],
            "prev_block": bd["prev_block"],
            "time": int(
                datetime.strptime(bd["time"], "%Y-%m-%dT%H:%M:%SZ")
                .replace(tzinfo=timezone.utc)
                .timestamp()
            ),
            "tx_count": bd["n_tx"],
            "txs": txs,
            "version": bd["ver"],
            "page": page,
            "pages": None
            if not limit
            else int(bd["n_tx"] // limit) + (bd["n_tx"] % limit > 0),
            "limit": limit,
        }
        return block

    # def getrawblock(self, blockid):

    def isspent(self, txid, output_n):
        t = self.gettransaction(txid)
        return 1 if t.outputs[output_n].spent else 0

    # def getinfo(self):
